#!/usr/bin/env sage-bootstrap-python


TEMPLATE = """
SAGE_LOCAL={SAGE_LOCAL}
prefix=${{SAGE_LOCAL}}
libdir=${{prefix}}/lib
includedir=${{prefix}}/include
Name: {target}
Version: {version}
Description: {target} for SageMath, provided by the OpenBLAS package.
Cflags: -I${{includedir}}
Libs: -L${{libdir}} {libflags}
""".lstrip()


import os


try:
    SAGE_LOCAL = os.environ['SAGE_LOCAL']
except KeyError:
    raise RuntimeError('must be run in a sage shell')

SAGE_DESTDIR = os.environ.get('SAGE_DESTDIR', '')


pkgconfigdir = os.path.join(SAGE_DESTDIR, SAGE_LOCAL.lstrip('/'), 'lib',
                            'pkgconfig')
if not os.path.isdir(pkgconfigdir):
    os.makedirs(pkgconfigdir)


with open('package-version.txt') as f:
    package_version = f.read()


def write_pc_file(target, libs, version):
    filename = os.path.join(pkgconfigdir, '{0}.pc'.format(target))
    libflags=' '.join('-l{0}'.format(lib) for lib in libs)
    pc_file = TEMPLATE.format(
        SAGE_LOCAL=SAGE_LOCAL,
        target=target,
        libflags=libflags,
        version=version,
    )
    with open(filename, 'w') as f:
        f.write(pc_file)
    print('Wrote {0}'.format(filename))


write_pc_file(  'blas', libs=['openblas'], version=package_version)
write_pc_file( 'cblas', libs=['openblas'], version=package_version)
write_pc_file('lapack', libs=['openblas'], version=package_version)




from sage_bootstrap.config import Configuration
config = Configuration()

from sage_bootstrap.stdio import init_streams
init_streams(config)

from sage_bootstrap.logger import init_logger
init_logger(config)



# -*- coding: utf-8 -*-
"""
PyPi Version Information
"""


# ****************************************************************************
#       Copyright (C) 2016      Volker Braun <vbraun.name@gmail.com>
#                     2020-2023 Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import logging
log = logging.getLogger()

import json

from sage_bootstrap.package import Package
from sage_bootstrap.updater import PackageUpdater
from sage_bootstrap.compat import urllib


class PyPiNotFound(Exception):
    pass


class PyPiError(Exception):
    pass


class PyPiVersion(object):

    def __init__(self, package_name, source='normal'):
        self.name = package_name
        self.json = self._get_json()
        # Replace provided name with the canonical name
        self.name = self.json['info']['name']
        if source == 'wheel':
            self.python_version = 'py3'
        else:
            self.python_version = 'source'

    def _get_json(self):
        response = urllib.urlopen(self.json_url)
        if response.getcode() != 200:
            raise PyPiNotFound('%s not on pypi', self.name)
        data = response.read()
        text = data.decode('utf-8')
        return json.loads(text)

    @property
    def json_url(self):
        return 'https://pypi.python.org/pypi/{0}/json'.format(self.name)

    @property
    def version(self):
        """
        Return the current version
        """
        return self.json['info']['version']

    @property
    def url(self):
        """
        Return the source url
        """
        for download in self.json['urls']:
            if self.python_version in download['python_version']:
                self.python_version = download['python_version']
                return download['url']
        raise PyPiError('No %s url for %s found', self.python_version, self.name)

    @property
    def tarball(self):
        """
        Return the source tarball name
        """
        for download in self.json['urls']:
            if self.python_version in download['python_version']:
                self.python_version = download['python_version']
                return download['filename']
        raise PyPiError('No %s url for %s found', self.python_version, self.name)

    @property
    def package_url(self):
        """
        Return the package URL
        """
        return self.json['info']['package_url']

    @property
    def license(self):
        """
        Return the package license
        """
        return self.json['info']['license']

    @property
    def summary(self):
        """
        Return the package summary
        """
        return self.json['info']['summary']

    @property
    def requires_dist(self):
        """
        Return the dependencies
        """
        return self.json['info']['requires_dist']

    @property
    def requires_python(self):
        """
        Return the requires_python attribute
        """
        return self.json['info']['requires_python']

    def update(self, package=None):
        if package is None:
            package = Package(self.name)
        if package.version == self.version:
            log.info('%s is already at the latest version', self.name)
            return
        log.info('Updating %s: %s -> %s', package.name, package.version, self.version)
        update = PackageUpdater(package.name, self.version)
        update.download_upstream(self.url)
        update.fix_checksum()



# -*- coding: utf-8 -*-
"""
Access the List of Sage Download Mirrors
"""

#*****************************************************************************
#       Copyright (C) 2014-2016 Volker Braun <vbraun.name@gmail.com>
#                     2015      Jeroen Demeyer
#                     2023      Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os
import contextlib
import logging
log = logging.getLogger()

from sage_bootstrap.compat import urllib, urlparse
from sage_bootstrap.env import SAGE_DISTFILES, SAGE_ROOT

from fcntl import flock, LOCK_SH, LOCK_EX
from errno import ENOLCK


def try_lock(fd, operation):
    """
    Try flock() but ignore ``ENOLCK`` errors, which could happen if the
    file system does not support locking.
    """
    try:
        flock(fd, operation)
    except IOError as e:
        if e.errno != ENOLCK:
            raise

        
class MirrorListException(RuntimeError):
    pass
        

class MirrorList(object):

    def __init__(self):
        self.sources = []
        upstream_d = os.path.join(SAGE_ROOT, '.upstream.d')
        for fname in sorted(os.listdir(upstream_d)):
            if '~' in fname or '#' in fname:
                # Ignore auto-save and backup files
                continue
            try:
                with open(os.path.join(upstream_d, fname), 'r') as f:
                    for line in f:
                        line = line.strip()
                        if line.startswith('#'):
                            continue
                        if not line:
                            continue
                        line = line.replace('${SAGE_ROOT}', SAGE_ROOT)
                        line = line.replace('${SAGE_DISTFILES}', SAGE_DISTFILES)
                        if '${SAGE_SERVER}' in line:
                            SAGE_SERVER = os.environ.get("SAGE_SERVER", "")
                            if not SAGE_SERVER:
                                continue
                            line = line.replace('${SAGE_SERVER}', SAGE_SERVER)
                        if line.endswith('mirror_list'):
                            cache_filename = os.path.join(SAGE_DISTFILES, line.rpartition('/')[2])
                            self.sources.append(MirrorList_from_url(line, cache_filename))
                        else:
                            self.sources.append([line])
            except IOError:
                # Silently ignore files that do not exist
                pass

    def __iter__(self):
        """
        Iterate through the list of mirrors.

        This is the main entry point into the mirror list. Every
        script should just use this function to try mirrors in order
        of preference. This will not just yield the official mirrors,
        but also urls for packages that are currently being tested.
        """
        for source in self.sources:
            for mirror in source:
                yield mirror


class MirrorList_from_url(object):
    
    MAXAGE = 24*60*60   # seconds

    def __init__(self, url, filename):
        self.url = url
        self.filename = filename
        self._mirrors = None

    @property
    def mirrors(self):
        if self._mirrors is not None:
            return self._mirrors

        try:
            self.mirrorfile = open(self.filename, 'r+t')
        except IOError:
            self.mirrorfile = open(self.filename, 'w+t')

        with self.mirrorfile:
            self.mirrorfd = self.mirrorfile.fileno()
            try_lock(self.mirrorfd, LOCK_SH)  # shared (read) lock
            if self._must_refresh():
                try_lock(self.mirrorfd, LOCK_EX)  # exclusive (write) lock
                # Maybe the mirror list file was updated by a different
                # process while we waited for the lock?  Check again.
                if self._must_refresh():
                    self._refresh()
            if self._mirrors is None:
                self._mirrors = self._load()

        return self._mirrors

    def _load(self, mirror_list=None):
        """
        Load and return `mirror_list` (defaults to the one on disk) as
        a list of strings
        """
        if mirror_list is None:
            try:
                self.mirrorfile.seek(0)
                mirror_list = self.mirrorfile.read()
            except IOError:
                log.critical('Failed to load the cached mirror list')
                return []
        if mirror_list == '':
            return []
        import ast
        try:
            return ast.literal_eval(mirror_list)
        except SyntaxError:
            log.critical('Downloaded mirror list has syntax error: {0}'.format(mirror_list))
            return []

    def _save(self):
        """
        Save the mirror list for (short-term) future  use.
        """
        self.mirrorfile.seek(0)
        self.mirrorfile.write(repr(self.mirrors))
        self.mirrorfile.truncate()
        self.mirrorfile.flush()

    def _port_of_mirror(self, mirror):
        if mirror.startswith('http://'):
            return 80
        if mirror.startswith('https://'):
            return 443
        if mirror.startswith('ftp://'):
            return 21
        # Sensible default (invalid mirror?)
        return 80

    def _rank_mirrors(self):
        """
        Sort the mirrors by speed, fastest being first

        This method is used by the YUM fastestmirror plugin
        """
        timed_mirrors = []
        import time
        import socket
        log.info('Searching fastest mirror')
        timeout = 1
        for mirror in self.mirrors:
            if not mirror.startswith('http'):
                log.debug('we currently can only handle http, got %s', mirror)
                continue
            port = self._port_of_mirror(mirror)
            mirror_hostname = urlparse.urlsplit(mirror).netloc
            time_before = time.time()
            try:
                sock = socket.create_connection((mirror_hostname, port), timeout)
                sock.close()
            except (IOError, socket.error, socket.timeout) as err:
                log.warning(str(err).strip() + ': ' + mirror)
                continue
            result = time.time() - time_before
            result_ms = int(1000 * result)
            log.info(str(result_ms).rjust(5) + 'ms: ' + mirror)
            timed_mirrors.append((result, mirror))
            timed_mirrors.sort()
            if len(timed_mirrors) >= 5 and timed_mirrors[4][0] < 0.3:
                # We don't need more than 5 decent mirrors
                break

        if len(timed_mirrors) == 0:
            # We cannot reach any mirror directly, most likely firewall issue
            if 'http_proxy' not in os.environ:
                log.error('Could not reach any mirror directly and no proxy set')
                raise MirrorListException('Failed to connect to any mirror, probably no internet connection')
            log.info('Cannot time mirrors via proxy, using default order')
        else:
            self._mirrors = [m[1] for m in timed_mirrors]
        log.info('Fastest mirror: ' + self.fastest)

    def _age(self):
        """
        Return the age of the cached mirror list in seconds
        """
        import time
        mtime = os.fstat(self.mirrorfd).st_mtime
        now = time.mktime(time.localtime())
        return now - mtime

    def _must_refresh(self):
        """
        Return whether we must download the mirror list.

        If and only if this method returns ``False`` is it admissible
        to use the cached mirror list.
        """
        if os.fstat(self.mirrorfd).st_size == 0:
            return True
        return self._age() > self.MAXAGE

    def _refresh(self):
        """
        Download and rank the mirror list.
        """
        log.info('Downloading the Sage mirror list')
        try:
            with contextlib.closing(urllib.urlopen(self.url)) as f:
                mirror_list = f.read().decode("ascii")
        except IOError:
            log.critical('Downloading the mirror list failed, using cached version')
        else:
            self._mirrors = self._load(mirror_list)
            self._rank_mirrors()
            self._save()

    def __iter__(self):
        """
        Iterate through the list of mirrors.

        This is the main entry point into the mirror list. Every
        script should just use this function to try mirrors in order
        of preference. This will not just yield the official mirrors,
        but also urls for packages that are currently being tested.
        """
        try:
            yield os.environ['SAGE_SERVER']
        except KeyError:
            pass
        for mirror in self.mirrors:
            if not mirror.endswith('/'):
                mirror += '/'
            yield mirror + '/'.join(['spkg', 'upstream', '${SPKG}'])

    @property
    def fastest(self):
        return next(iter(self))



# -*- coding: utf-8 -*-
"""
Interface to the Sage fileserver
"""

# ****************************************************************************
#       Copyright (C) 2016 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import subprocess


class FileServer(object):

    def __init__(self):
        self.user = 'sagemath'
        self.hostname = 'fileserver.sagemath.org'

    def upstream_directory(self, package):
        """
        Return the directory where the tarball resides on the server
        """
        return os.path.join(
            '/', 'data', 'files', 'spkg', 'upstream', package.name,
        )

    def upload(self, package):
        """
        Upload the current tarball of package
        """
        if not package.tarball.is_distributable():
            raise ValueError('Tarball of {} is marked as not distributable'.format(package))
        subprocess.check_call([
            'ssh', 'sagemath@fileserver.sagemath.org',
            'mkdir -p {0} && touch {0}/index.html'.format(self.upstream_directory(package))
        ])
        subprocess.check_call([
            'rsync', '-av', '--checksum', '-e', 'ssh -l sagemath',
            package.tarball.upstream_fqn,
            'fileserver.sagemath.org:{0}'.format(self.upstream_directory(package))
        ])

    def publish(self):
        """
        Publish the files
        """
        subprocess.check_call([
            'ssh', 'sagemath@fileserver.sagemath.org', './publish-files.sh'
        ])
